"""accelpy Fortran-based Accelerator module"""

import os

from collections import OrderedDict

from accelpy.util import get_f_dtype
from accelpy.accel import AccelBase


moddatasrc = """
module {datamodname}
USE, INTRINSIC :: ISO_C_BINDING

{use}

public dataenter_{runid}, dataexit_{runid}

contains

INTEGER (C_INT64_T) FUNCTION dataenter_{runid}({enterargs}) BIND(C, name="dataenter_{runid}")
    USE, INTRINSIC :: ISO_C_BINDING

    {entervardefs}

    {enterdirective}

    dataenter_{runid} = 0

END FUNCTION

INTEGER (C_INT64_T) FUNCTION dataexit_{runid}({exitargs}) BIND(C, name="dataexit_{runid}")
    USE, INTRINSIC :: ISO_C_BINDING

    {exitvardefs}

    {exitdirective}

    dataexit_{runid} = 0

END FUNCTION

end module
"""

modkernelsrc = """
INTEGER (C_INT64_T) FUNCTION runkernel_{runid}({kernelargs}) BIND(C, name="runkernel_{runid}")
    USE, INTRINSIC :: ISO_C_BINDING

    {use}

    {kernelvardefs}

    {kernelbody}

    runkernel_{runid} = 0

END FUNCTION
"""


class FortranAccelBase(AccelBase):

    lang = "fortran"
    srcext = ".F90"

    @classmethod
    def _gen_include(cls):
        return []

    @classmethod
    def _dimension(cls, arg, attrspec):

        aid = arg["id"]

        if aid in attrspec and "dimension" in attrspec[aid]:
            return attrspec[aid]["dimension"]
                
        return ", ".join([str(s) for s in arg["data"].shape])

    @classmethod
    def _modvardefs(cls, arg):

        dim =  ", ".join([":" for s in arg["data"].shape])

        if dim:
            return "%s, DIMENSION(%s), POINTER :: %s" % (get_f_dtype(arg),
                dim, arg["curname"])

        else:
            return "%s, POINTER :: %s" % (get_f_dtype(arg), arg["curname"])

    @classmethod
    def _funcvardefs(cls, arg, intent, attrspec={}):

        dim = cls._dimension(arg, attrspec)

        if dim:
            return "%s, DIMENSION(%s), INTENT(%s), TARGET :: %s" % (get_f_dtype(arg),
                dim, intent, arg["curname"])

        else:
            return "%s, INTENT(IN) :: %s" % (get_f_dtype(arg), arg["curname"])
#
#    @classmethod
#    def _kernelvardefs(cls, arg, intent, attrspec={}):
#
#        curname = arg["curname"]
#
#        if curname in attrspec and "dimension" in attrspec[curname]:
#            dim = attrspec[curname]["dimension"]
#
#        else:
#            dim = ", ".join([str(s) for s in arg["data"].shape])
#
#        if dim:
#            return "%s, DIMENSION(%s), INTENT(%s), TARGET :: %s" % (get_f_dtype(arg),
#                dim, intent, arg["curname"])
#
#        else:
#            return "%s, INTENT(IN) :: %s" % (get_f_dtype(arg), arg["curname"])

    @classmethod
    def gen_kernelfile(cls, knlhash, dmodname, runid, specid, modattr, section, workdir, localvars, modvars):

        kernelpath = os.path.join(workdir, "K%s%s" % (knlhash[2:], cls.srcext))

        kernelparams = {
            "runid": str(runid) + str(specid)
        }

        kernelargs = []
        kernelvardefs = []

        attrspec = section.kwargs.get("attrspec", {})
        modattr.update(attrspec)

        for old, mvar in modvars:

            kernelargs.append(mvar["curname"])
            kernelvardefs.append(cls._funcvardefs(mvar, "INOUT", attrspec=modattr))

        for lvar in localvars:

            kernelargs.append(lvar["curname"])
            kernelvardefs.append(cls._funcvardefs(lvar, "INOUT", attrspec=attrspec))

        kernelparams["kernelmodname"] = "MOD%s" % knlhash[2:].upper()
        kernelparams["kernelargs"] = ", ".join(kernelargs)
        kernelparams["kernelvardefs"] = "\n".join(kernelvardefs)
        kernelparams["kernelbody"] = "\n".join(section.body)
        kernelparams["use"] = "\n".join(cls._gen_include())

        with open(kernelpath, "w") as fkernel:
            fkernel.write(modkernelsrc.format(**kernelparams))

        #import pdb; pdb.set_trace()
        return kernelpath

    @classmethod
    def gen_datafile(cls, modname, filename, runid, workdir, copyinout, copyin, copyout, alloc, attr):

        datapath = os.path.join(workdir, filename)

        dataparams = {"runid": str(runid), "datamodname": modname}

        enterargs = []
        entervardefs = []
        exitargs = []
        exitvardefs = []

        enterdirective = []
        exitdirective = []

        alnames = []

        cionames = []

        for cio in copyinout:
            cioname = cio["curname"]
            cionames.append(cioname)

            enterargs.append(cioname)
            entervardefs.append(cls._funcvardefs(cio, "INOUT", attrspec=attr))

            exitargs.append(cioname)
            exitvardefs.append(cls._funcvardefs(cio, "INOUT", attrspec=attr))

        if cionames:
            enterdirective.append(cls._mapto(cionames))
            exitdirective.append(cls._mapfrom(cionames))

        cinames = []

        for ci in copyin:
            ciname = ci["curname"]
            cinames.append(ciname)

            enterargs.append(ciname)
            entervardefs.append(cls._funcvardefs(ci, "INOUT", attrspec=attr))

        if cinames:
            enterdirective.append(cls._mapto(cinames))

        conames = []

        for co in copyout:
            coname = co["curname"]
            conames.append(coname)

            enterargs.append(coname)
            entervardefs.append(cls._funcvardefs(co, "INOUT", attrspec=attr))

            exitargs.append(coname)
            exitvardefs.append(cls._funcvardefs(co, "INOUT", attrspec=attr))


        if conames:
            alnames.extend(conames)
            exitdirective.append(cls._mapfrom(conames))

        for al in alloc:
            alname = al["curname"]
            alnames.append(alname)

            enterargs.append(alname)
            entervardefs.append(cls._funcvardefs(al, "INOUT", attrspec=attr))

        if alnames:
            enterdirective.append(cls._mapalloc(alnames))

        dataparams["enterargs"] = ", ".join(enterargs)
        dataparams["entervardefs"] = "\n".join(entervardefs)
        dataparams["enterdirective"] = "\n".join(enterdirective)
        dataparams["exitargs"] = ", ".join(exitargs)
        dataparams["exitvardefs"] = "\n".join(exitvardefs)
        dataparams["exitdirective"] = "\n".join(exitdirective)
        dataparams["use"] = "\n".join(cls._gen_include())

        with open(datapath, "w") as fdata:
            fdata.write(moddatasrc.format(**dataparams))

        #import pdb; pdb.set_trace()
        return datapath

class FortranAccel(FortranAccelBase):
    accel = "fortran"

    @classmethod
    def _mapto(cls, names):
        return ""

    @classmethod
    def _mapfrom(cls, names):
        return ""

    @classmethod
    def _mapalloc(cls, names):
        return ""


class OpenmpFortranAccel(FortranAccel):
    accel = "openmp"

    @classmethod
    def _gen_include(cls):
        return ["USE OMP_LIB"]


class AcctargetFortranAccel(FortranAccelBase):
    pass


class OmptargetFortranAccel(AcctargetFortranAccel):
    accel = "omptarget"

    @classmethod
    def _mapto(cls, names):
        return "!$omp target enter data map(to:" + ", ".join(names) + ")"

    @classmethod
    def _mapfrom(cls, names):
        return "!$omp target exit data map(from:" + ", ".join(names) + ")"

    @classmethod
    def _mapalloc(cls, names):
        return "!$omp target enter data map(alloc:" + ", ".join(names) + ")"


class OpenaccFortranAccel(AcctargetFortranAccel):
    accel = "openacc"

    @classmethod
    def _mapto(cls, names):
        return "!$acc enter data copyin(" + ", ".join(names) + ")"

    @classmethod
    def _mapfrom(cls, names):
        return "!$acc exit data copyout(" + ", ".join(names) + ")"

    @classmethod
    def _mapalloc(cls, names):
        return "!$acc enter data create(" + ", ".join(names) + ")"

_fortaccels = OrderedDict()
AccelBase.avails[FortranAccelBase.lang] = _fortaccels

_fortaccels[OmptargetFortranAccel.accel] = OmptargetFortranAccel
_fortaccels[OpenaccFortranAccel.accel] = OpenaccFortranAccel
_fortaccels[OpenmpFortranAccel.accel] = OpenmpFortranAccel
_fortaccels[FortranAccel.accel] = FortranAccel




"""accelpy Kernel module"""


import os, abc, time, threading, shutil, numpy, tempfile

from ctypes import c_int64
from numpy.ctypeslib import ndpointer, load_library
from collections import OrderedDict

from accelpy.const import (version, NOCACHE, MEMCACHE, FILECACHE, NODEBUG,
                            MINDEBUG, MAXDEBUG, NOPROF, MINPROF, MAXPROF,
                            accel_priority)
from accelpy.util import (Object, gethash, get_config, set_config,
                            pack_arguments, getname_varmap)
from accelpy.spec import Spec
from accelpy.compiler import Compiler, get_compilers
from accelpy.cache import slib_cache


class Task(Object):

    def __init__(self, liblang, libpath, bldpath):

        self.liblang = liblang
        self.libpath = libpath
        self.bldpath = bldpath
        self.synched = False

#    def __enter__(self, *data, specenv={}, timeout=None):
#        self.timeout = timeout
#        self.launch(*data, specenv=specenv)
#
#        return self
#
#    def __exit__(self, *exc):
#        self.wait(timeout = self.timeout)
#
#        return False

    def run(self, data):

        self.thread = threading.Thread(target=self._start_kernel,
                            args=(data,))
        self.start = time.time()
        self.thread.start()

    def varmap(self, arg, funcname):

        if self.liblang[1] == "cpp":
            flags = ["c_contiguous"]

        elif self.liblang[1] == "fortran":
            flags = ["f_contiguous"]

        else:
            raise Exception("Unknown language: %s" % self.liblang[1])

        # TODO consider to use "numpy.ascontiguousarray"

        datamap = getattr(self.liblang[0], funcname)
        datamap.restype = c_int64
        datamap.argtypes = [ndpointer(arg["data"].dtype, flags=",".join(flags))]

        res = datamap(arg["data"])

        return res

    def _start_kernel(self, data):

        lib = None

        if self.liblang[0] is None:
            if self.libpath is not None and os.path.isfile(self.libpath):
                try:
                    libdir, libname = os.path.split(self.libpath)
                    basename, _ = os.path.splitext(libname)

                    lib = load_library(basename, libdir)
                except:
                    pass

            if (lib is None and self.bldpath is not None and
                    os.path.isfile(self.bldpath)):
                blddir, bldname = os.path.split(self.bldpath)
                basename, _ = os.path.splitext(bldname)

                lib = load_library(basename, blddir)

        if lib is not None:
            with threading.Lock():
                self.liblang[0] = lib

        for arg in data:
            self.varmap(arg, getname_varmap(arg))

        start = getattr(self.liblang[0], "accelpy_start")
        start.restype = c_int64
        start.argtypes = []

        res = start()

        if res != 0:
            raise Exception("kernel returns non-zero: %d" % res)

    def wait(self, timeout=None):

        timeout = max(0, self.start+timeout-time.time()) if timeout else None

        while self.thread.ident is None:
            time.sleep(0.1)
            
        if self.thread.is_alive():
            self.thread.join(timeout=timeout)

        if (self.libpath and self.bldpath and not os.path.isfile(self.libpath)
            and os.path.isfile(self.bldpath)):
            try:
                shutil.copyfile(self.bldpath, self.libpath)

            except FileExistsError:
                pass

        self.synched = True

    def stop(self, timeout=None):

        if not self.synched:
            self.wait(timeout=timeout)

        accstop = getattr(self.liblang[0], "accelpy_stop")
        accstop.restype = c_int64
        accstop.argtypes = []

        res = accstop()

        return res


class KernelBase(Object):

    avails = OrderedDict()

    #def __init__(self, spec, compile=None, debug=NODEBUG):
    def __init__(self, section, data, cache=MEMCACHE, profile=NOPROF, debug=NODEBUG):

        self.section = section
        self.data = data
        self.debug = debug
        self.cache = cache
        self.profile = profile

#        self.spec = spec if isinstance(spec, Spec) else Spec(spec)
#        self.compile = compile
#        self.cachekey = None
#        self.liblang = [None, None]
#
#        if self.spec is None:
#            raise Exception("No kernel spec is found")

    @abc.abstractmethod
    def get_dtype(self, arg):
        pass

    def get_include(self):
        return ""

    def add_includes(self):

        incs = []

        # TODO: implement this
        #import pdb; pdb.set_trace()

        return incs

    @abc.abstractmethod
    def gen_code(self, compiler):
        pass


class Kernel:

    def __init__(self, spec, acctype=None, compile=None, cache=MEMCACHE, profile=0, debug=NODEBUG):

        self.acctype = acctype
        self.name = None
        self._kernel = None

        self.cache = cache
        self.profile = profile
        self.debug = debug
        self.compile = compile
        self.cachekey = None
        self.liblang = [None, None]

        if isinstance(spec, Spec):
            self.spec = spec
        elif spec is not None:
            self.spec = Spec(spec)
        else:
            raise Exception("No kernel spec is found")

        self.tasks = []

    def set_name(self, name):

        if self.acctype is None:
            self.name = name

        elif self.acctype == name:
            self.name = name

        elif isinstance(self.acctype, (list, tuple)) and name in self.acctype:
            self.name = name

        else:
            raise Exception("Can not set Kernel name: (%s | %s)" %
                            (str(self.acctype), name))

    def launch(self, *data, environ={}):

        self.spec.eval_pysection(environ)
        self.data = pack_arguments(data)

        if self.name is None or self._kernel is None:
            self.set_kernel()

        else:
            self.section = self.spec.get_section(self.name)

        if self._kernel is None:
            if self.name is None:
                raise Exception("Kernel type is not defined.")

            else:
                self._kernel = KernelBase.avails[self.name](self.section, self.data)

        if self.debug > 0:
            print("DEBUG: using '%s' kernel" % (self.name))

        self.spec.update_argnames(self.data)
        self.section.update_argnames(self.data)

        keys = [os.uname().release, version, self.name,
                self.compile, self.section.hash()]

        for item in self.data:
            keys.append(item["data"].shape)
            keys.append(item["data"].dtype)

        if self.liblang[0] is None or self.cache < MEMCACHE:
            lang, libpath, bldpath = self.build_sharedlib(gethash(str(keys)))
            self.liblang[0] = None
            self.liblang[1] = lang
            task = Task(self.liblang, libpath, bldpath)

        else:
            task = Task(self.liblang, None, None)

        task.run(self.data)

        self.tasks.append(task)

        return task

    def wait(self, *tasks, timeout=None):

        if len(tasks) > 0:
            for task in tasks:
                task.wait(timeout=timeout)

        else:
            for task in self.tasks:
                task.wait(timeout=timeout)

    def stop(self, timeout=None):

        for task in self.tasks:
            if not task.synched:
                task.wait(timeout=timeout)

            task.stop(timeout=timeout)

    def build_sharedlib(self, ckey):

        errmsgs = []

        compilers = get_compilers(self.name, self._kernel.lang,
                        compile=self.compile, debug=self.debug)

        for comp in compilers:

            cachekey = ckey + "_" + comp.hashkey

            if self.cache >= FILECACHE and cachekey in slib_cache:
                lang, basename, libext, libpath, bldpath = slib_cache[cachekey]
                self.cachekey = cachekey

                return lang, libpath, bldpath

            libdir = os.path.join(get_config("libdir"), comp.vendor, cachekey[:2])

            if not os.path.isdir(libdir):
                try:
                    os.makedirs(libdir)

                except FileExistsError:
                    pass

            basename = cachekey[2:]
            libname = basename + "." + comp.libext
            libpath = os.path.join(libdir, libname)

            if self.cache >= FILECACHE and os.path.isfile(libpath):
                self.cachekey = cachekey
                slib_cache[self.cachekey] = (comp.lang, basename, comp.libext,
                                                    libpath, None)
                return comp.lang, libpath, None

            try:
                codes, macros = self._kernel.gen_code(comp)

                if not os.path.isdir(get_config("blddir")):
                    set_config("blddir", tempfile.mkdtemp())

                bldpath = comp.compile(codes, macros)

                self.cachekey = cachekey
                slib_cache[self.cachekey] = (comp.lang, basename, comp.libext,
                                                    libpath, bldpath)
                return comp.lang, libpath, bldpath

            except Exception as err:
                errmsgs.append(str(err))

        raise Exception("\n".join(errmsgs))


    def set_kernel(self, acctype=None):

        if isinstance(self.acctype, str):
            acctype = (self.acctype,)

        accel = set(KernelBase.avails.keys()) if acctype is None else set(acctype)
        accel &= set(self.spec.list_sections())

        errmsgs = []

        for a in sorted(list(accel), key=(lambda x: accel_priority.index(x))):
            try:
                section = self.spec.get_section(a)
                self._kernel = KernelBase.avails[a](section, self.data, debug=self.debug)
                self.name = self._kernel.name
                self.section = section

                return

            except Exception as err:
                errmsgs.append(repr(err))

        raise Exception("No kernel is available: %s" % "\n".join(errmsgs))


#def Kernel(spec, accel=None, compile=None, cache=MEMCACHE, profile=0, debug=NODEBUG):
#
#    if isinstance(accel, str):
#        return KernelBase.avails[accel](spec, compile=compile, cache=cache,
#                                        profile=profile, debug=debug)
#
#    elif accel is None or isinstance(accel, (list, tuple)):
#
#        if accel is None:
#            accel = KernelBase.avails.keys()
#
#        errmsgs = []
#
#        for k in accel:
#            try:
#                kernel = KernelBase.avails[k](spec, compile=compile, cache=cache,
#                                                profile=profile, debug=debug)
#                return kernel
#
#            except Exception as err:
#                errmsgs.append(repr(err))
#
#        raise Exception("No kernel is available: %s" % "\n".join(errmsgs))
#
#    raise Exception("Kernel '%s' is not valid." % str(accel))
#
#
#def generate_compiler(compile):
#
#    clist = compile.split()
#
#    compcls = Compiler.from_path(clist[0])
#    comp = compcls(option=" ".join(clist[1:]))
#
#    return comp
#
#
#def get_compilers(kernel, compile=None):
#    """
#        parameters:
#
#        kernel: the kernel id or a list of them
#        compile: compiler path or compiler id, or a list of them
#                  syntax: "vendor|path [{default}] [additional options]"
#"""
#
#    kernels = []
#
#    if isinstance(kernel, str):
#        kernels.append((kernel, KernelBase.avails[kernel].lang))
#
#    elif isinstance(kernel, KernelBase):
#        kernels.append((kernel.name, kernel.lang))
#
#    elif isinstance(kernel, (list, tuple)):
#        for k in kernel:
#            if isinstance(k, str):
#                kernels.append((k, KernelBase.avails[k].lang))
#
#            elif isinstance(k, KernelBase):
#                kernels.append((k.name, k.lang))
#
#            else:
#                raise Exception("Unknown kernel type: %s" % str(k))
#
#    else:
#        raise Exception("Unknown kernel type: %s" % str(kernel))
#
#    compilers = []
#
#    if compile:
#        if isinstance(compile, str):
#            citems = compile.split()
#
#            if not citems:
#                raise Exception("Blank compile")
#
#            if os.path.isfile(citems[0]):
#                compilers.append(generate_compiler(compile))
#
#            else:
#                # TODO: vendor name search
#                for lang, langsubc in Compiler.avails.items():
#                    for kernel, kernelsubc in langsubc.items():
#                        for vendor, vendorsubc in kernelsubc.items():
#                            if vendor == citems[0]:
#                                try:
#                                    compilers.append(vendorsubc(option=" ".join(citems[1:])))
#                                except:
#                                    pass
#
#        elif isinstance(compile, Compiler):
#            compilers.append(compile)
#
#        elif isinstance(compile, (list, tuple)):
#
#            for comp in compile:
#                if isinstance(comp, str):
#                    citems = comp.split()
#
#                    if not citems:
#                        raise Exception("Blank compile")
#
#                    if os.path.isfile(citems[0]):
#                        try:
#                            compilers.append(generate_compiler(comp))
#                        except:
#                            pass
#
#                    else:
#                        # TODO: vendor name search
#                        for lang, langsubc in Compiler.avails.items():
#                            for kernel, kernelsubc in langsubc.items():
#                                for vendor, vendorsubc in kernelsubc.items():
#                                    if vendor == citems[0]:
#                                        try:
#                                            compilers.append(vendorsubc(option=" ".join(citems[1:])))
#                                        except:
#                                            pass
#
#                elif isinstance(comp, Compiler):
#                    compilers.append(comp)
#
#                else:
#                    raise Exception("Unsupported compiler type: %s" % str(comp))
#        else:
#            raise Exception("Unsupported compiler type: %s" % str(compile))
#
#    return_compilers = []
#    errmsgs = []
#
#
#    if compilers:
#        for comp in compilers:
#            if any(comp.accel==k[0] and comp.lang==k[1] for k in kernels):
#                return_compilers.append(comp)
#
#    elif compile is None:
#        for acc, lang in kernels:
#
#            if lang not in Compiler.avails:
#                continue
#
#            if acc not in Compiler.avails[lang]:
#                continue
#
#            vendors = Compiler.avails[lang][acc]
#
#            for vendor, cls in vendors.items():
#                try:
#                    return_compilers.append(cls())
#                except Exception as err:
#                    errmsgs.append(str(err))
#
#
#    if not return_compilers:
#        if errmsgs:
#            raise Exception("No compiler is found: %s" % "\n".join(errmsgs))
#        else:
#            raise Exception("No compiler is found: %s" % str(kernels))
#
#    return return_compilers
#



"""accelpy Spec module"""

import os, hashlib

from accelpy.util import Object, _accelpy_builtins, appeval, funcargseval, gethash


class Section():

    def __init__(self, accel, vargs, kwargs, body, env):
        self.accel = accel
        self.argnames = vargs
        self._kwargs = kwargs
        self.kwargs = None
        self.body = body
        self.env = env
        self.md5 = None

    def hash(self):

        if self.md5 is None:
            self.md5 = gethash("".join(self.body))

        return self.md5

    def is_enabled(self):
        return self.kwargs.get("enable", True)

    def kind(self):
        return self.accel 

    def update_argnames(self, data):

        for idx, arg in enumerate(data):

            if self.argnames is not None and len(self.argnames) > idx:
                arg["curname"] = self.argnames[idx]


class Spec(Object):

    def __init__(self, spec):

        # invargs, outvars, kwargs
        self._argnames = None

        if isinstance(spec, str):
            if os.path.isfile(spec):
                with open(spec) as fs:
                    spec = Spec(fs.read())

            self._sections = self._parse_spec(spec)

        elif isinstance(spec, Spec):
            self._sections = spec._sections 

        else:
            raise Exception("Wrong spec type: %s" % str(spec))

    def _of_set_argnames(self, *vargs):

        self._argnames = list(vargs)

    def _parse_spec(self, spec):

        rawlines = spec.split("\n")

        sec_starts = []

        for lineno, rawline in enumerate(rawlines):
            if rawline and rawline[0] == "[":
                    sec_starts.append(lineno)

        if len(sec_starts) == 0:
            raise Exception("No spec is found.")

        sec_starts.append(len(rawlines))

        self._env = dict(_accelpy_builtins)
        self._env["set_argnames"] =  self._of_set_argnames

        self.env = self._env.copy()

        self._pysection = rawlines[0:sec_starts[0]]

        sections = []
        for sec_start, sec_end in zip(sec_starts[0:-1], sec_starts[1:]):
            section = self._parse_section(rawlines[sec_start:sec_end])
            sections.extend(section)

        return sections

    def eval_pysection(self, environ):

        self.env = self._env.copy()

        if isinstance(environ, dict):
            self.env.update(environ)

        _, lenv = appeval("\n".join(self._pysection), self.env)

        self.env.update(lenv)

    def _parse_section(self, rawlines):

        assert (rawlines[0] and rawlines[0][0] == "[")

        maxlineno = len(rawlines)
        row = 0
        col = 1

        names = []
        arg_start = None

        # collect accelerator names
        while(row < maxlineno):

            if rawlines[row].lstrip().startswith("#"):
                row += 1
                col = 0
                continue

            for idx in range(col, len(rawlines[row])):
                c = rawlines[row][idx]
                if c in (":", "]"):
                    names.append(rawlines[row][col:idx])
                    arg_start = [c, row, idx+1]
                    row = maxlineno
                    break
            if row < maxlineno:
                names.append(rawlines[row])
            row += 1
            col = 0

        assert names

        accels = [n.strip() for n in "".join(names).split(",")]
        
        args = []

        char, row, col = arg_start

        # collect accelerator arguments
        if char != "]":

            while(row < maxlineno):

                line = rawlines[row][col:].rstrip()

                if not line or line[-1] != "]":
                    args.append(line)
                    row += 1
                    col = 0
                    continue
                else:
                    args.append(line[:-1])

                try:
                    fargs = " ".join(args).split(",")
                    vargs = [a.strip() for a in fargs if "=" not in a]
                    kwargs = [a.strip() for a in fargs[len(vargs):]]
                    body = rawlines[row+1:]
                    break

                except Exception as err:
                    row += 1
                    col = 0
        else:
            vargs, kwargs, body = [], {}, rawlines[row+1:]

        sections = []

        for accel in accels:
            sections.append(Section(accel, vargs, kwargs, body, self.env))

        return sections

    def update_argnames(self, data):

        for idx, arg in enumerate(data):

            if self._argnames is not None and len(self._argnames) > idx:
                arg["curname"] = self._argnames[idx]

            else:
                arg["curname"] = "accpy_var%d" % idx

    def get_section(self, accel):

        for sec in self._sections:

            sec.kwargs = funcargseval(",".join(sec._kwargs), self.env)

            if sec.accel == accel and sec.is_enabled():
                return sec

        raise Exception("No section with '%s' exists or is enabled." % accel)

    def list_sections(self, acctype=None):

        if isinstance(acctype, str):
            acctype = (acctype,)

        secs = []

        for sec in self._sections:

            sec.kwargs = funcargseval(",".join(sec._kwargs), self.env)

            if sec.is_enabled():
                if acctype and sec.accel not in acctype:
                    continue

                secs.append(sec.accel)

        return secs
